#include <asm/unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <asm/signal.h>
#include <sys/time.h>
#include <asm/fcntl.h>

pid_t waitpid(pid_t, int *, int);
pid_t getpgid(pid_t);
pid_t fork(void);
int execve(const char *, char *const [], char *const []) __attribute__(( __noreturn__ ));
int sethostname(const char *, size_t);
int setdomainname(const char *, size_t);
int open(const char *, int);
int close(int);
int pause(void);
ssize_t read(int, void *, size_t);
ssize_t write(int, const void *, size_t);
void (*signal(int, void (*)(int)))(int);
void exit(int) __attribute__(( __noreturn__ ));
void chop(char *);

extern char str_autoopt;
extern char str_confpath;
extern char str_hostname;
extern char str_domainname;
extern char str_spawn;
extern char bannerfmt;
#ifdef VERBOSE
extern char respawnfmt;
#endif /* VERBOSE */
extern char nullstr;
extern unsigned long num;

#define MSGLEN		512
#define MAXCOMLEN	512
#define MAXOPTLEN	32
#define MAXARG		8
#define MAXENV		8
#define MAXSPAWNS	8

/*
 *  minimal printf
 *  buffer overflow safe
 */
void printf(const char *format, ...)
{
	char	tmp[MSGLEN];
	const	char **pp,*pt;
	const unsigned long *np;
	char	*src,*dest;
	unsigned long n;

	pp = &format;
	pp++;
	pt = format;
	dest = tmp;
	while(*pt)
	{
		switch(*pt)
		{
		case '%':
			pt++;
			switch(*pt)
			{
			case 's':
				pt++;
				src = (char*)*(pp++);
				if (!src)
					src = &nullstr;
				while(*src)
				{
					*(dest++) = *(src++);
				}
				break;
			case 'i':
				pt++;
				(int)n = (int)*(pp++);
				if ((int)n < 0)
				{
					(int)n = -((int)n);
					*(dest++) = '-';
				}
				*dest = '0';
				np = &num;
				for(;*np;np++)
				{
					if (n >= *np)
						goto non_zero;
				}
				for(;*np;np++)
				{
					while(n >= *np)
					{
				non_zero:
						n -= *np;
						(*dest)++;
					}
					dest++;
					*dest = '0';
				}
				*(dest++) = '0' + n;
				break;
			default:
				return;
			}
			break;
		default:
			*(dest++) = *(pt++);
		}
	}
	write(1,tmp,(int)(dest-tmp));
}

int run(const char *command, const int w)
{
	char	copy[strlen(command)+1];
	char	*a[MAXARG];
	char	*e[MAXENV];
	pid_t	pid;
	int	i;

	if ((pid = fork()) == 0)
	{
		a[0] = copy;
		a[1] = copy;
		i = 1;
		while(*command)
		{
			if (*command == ' ')
			{
				*(a[i]++) = 0;
				a[i+1] = a[i];
				i++;
				while(*command == ' ')
					command++;
			}
			else
			{
				*(a[i]++) = *(command++);
			}
		}
		a[i] = 0;
		a[i] = NULL;
		e[0] = NULL;
		execve(a[0],(char **)&a[1],(char **)e);
		/* NOT REACHED */
	}
	if (w)
	{
		/*
		 *  at this stage, size is more important
		 */
		waitpid(pid,NULL,0);
	}
	return(pid);
}

void atomic_optget(const char *path, char *buffer, int bufsize)
{
	int	n,fd;

	if ((fd = open(path,O_RDONLY)) < 0)
	{
		*buffer = 0;
		return;
	}
	n = read(fd,buffer,bufsize-1);
	buffer[n] = 0;
	chop(buffer);	/* remove \n and anything after */
	close(fd);
}

void sigchld(int sig)
{
	waitpid(-1,NULL,WNOHANG);
	signal(SIGCHLD,sigchld);
}

void glue(char *dest, const char *s1, const char *s2)
{
	while(*s1)
		*(dest++) = *(s1++);
	while(*s2)
		*(dest++) = *(s2++);
	*dest = 0;
}

int main(int argc, char **argv, char **envp) __attribute__(( __noreturn__ ));
int main(int argc, char **argv, char **envp)
{
	printf(&bannerfmt,argv[1]);

	/*
	 *  smaller than calling signal(SIGCHLD,sigchld)
	 */
	sigchld(0);

	/*
	 *  sethostname and setdomainname
	 */
	{
		char	name[128];
		char	*p;

		glue(name,&str_confpath,&str_hostname);
		atomic_optget(name,name,sizeof(name));
		if (name[0])
		{
			for(p=name;*p;p++)
				;
			sethostname(name,p-name);
		}
		glue(name,&str_confpath,&str_domainname);
		atomic_optget(name,name,sizeof(name));
		if (name[0])
		{
			for(p=name;*p;p++)
				;
			setdomainname(name,p-name);
		}
	}

	/*
	 *  figure out what to run
	 */
	{
		char	buf[MAXCOMLEN];

		glue(buf,&str_confpath,(argv[1]) ? argv[1] : &str_autoopt);
		atomic_optget(buf,buf,sizeof(buf));
		if (buf[0])
			run(buf,1);
	}
	/*
	 *  major kludge-alert
	 *  and no, this is not an entry for the obfuscated C contest...
	 *  (atleast not intended as such)
	 */
	{
		struct	respawn { pid_t pid; char run[MAXOPTLEN]; } spawn[MAXSPAWNS];
		int	x,ret;

		for(x=0;x<MAXSPAWNS;x++)
			spawn[x].pid = 1;

		/*
		 *  read a list of spawns
		 */
		{
			char	buf[MAXCOMLEN];
			char	*dest,*src;

			glue(buf,&str_confpath,&str_spawn);
			atomic_optget(buf,buf,sizeof(buf));
			if (buf[0])
			{
				x = 0;
				src = buf;
			nextspawn:
				dest = spawn[x].run;
				spawn[x].pid = -1;
				while(*src && *src != ' ')
				{
					*(dest++) = *(src++);
				}
				*dest = 0;
				if (*src == ' ')
				{
					x++;
					while(*src == ' ')
						src++;
					if (*src && (x < MAXSPAWNS))
						goto nextspawn;
				}
			}
		}
		goto first_check;
		/*
		 *  this is where we handle respawns and such...
		 */
loop:
		pause();
		/*
		 *  after each signal (SIGCHLD) we come here
		 */
first_check:
		for(x=0;x<MAXSPAWNS;x++)
		{
			ret = getpgid(spawn[x].pid);
			if (ret < 0)
			{
				char	buf[MAXCOMLEN];

				glue(buf,&str_confpath,spawn[x].run);
				atomic_optget(buf,buf,sizeof(buf));
				if (buf[0])
				{
					spawn[x].pid = run(buf,0);
					/*
					 *  buf is chopped in child only
					 */
#ifdef VERBOSE
					printf(&respawnfmt,buf,x+1,spawn[x].run,spawn[x].pid);
#endif /* VERBOSE */
				}
			}
		}
		goto loop;
	}
}

